/*
*   socket/linux.cpp is a part of programm carbon
*   $Id$
*   Copyright (C) 2005-2006  Matwey V. Kornilov
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <algorithm>
#include <memory>

#include "linux.h"
#include "socket.h"

IP4Addr& operator<<( IP4Addr& addr, sockaddr_in* sa_addr ){
    std::string str( inet_ntoa( sa_addr->sin_addr ) );
    addr = IP4Addr( str, sa_addr->sin_port );
    return addr;
}

const IP4Addr& operator>>( const IP4Addr& addr, sockaddr_in* sa_addr ){
    hostent *host = gethostbyname( addr.getName().c_str() );
    if( !host ) throw SocketError( hstrerror(h_errno) );
    sa_addr->sin_addr = *(struct in_addr *)host->h_addr; //  name.sin_addr.s_addr = htonl (INADDR_ANY);
    sa_addr->sin_port = htons(addr.getPort());
    sa_addr->sin_family = AF_INET;
    return addr;
}

template <class AddrT> LinuxSocketTraits<AddrT>::LinuxSocketTraits(){
    int ret = socket( PF_INET, SOCK_STREAM, 0 );
    if( ret < 0 ) throw SocketError( strerror(errno) , typename SocketError::Error(errno) );
    m_handle = ret;
    int so_ra = 1;
    if(::setsockopt( m_handle, SOL_SOCKET, SO_REUSEADDR, &so_ra, sizeof(int))<0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    if(::fcntl( m_handle, F_SETFL,::fcntl( m_handle, F_GETFL, 0)|O_NONBLOCK ) == -1 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
}

template <class AddrT> LinuxSocketTraits<AddrT>::LinuxSocketTraits( int handle ){
    m_handle = handle;
    if(::fcntl( m_handle, F_SETFL,::fcntl( m_handle, F_GETFL, 0)|O_NONBLOCK) == -1 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
}

template <class AddrT> LinuxSocketTraits<AddrT>::~LinuxSocketTraits(){
    ::close(m_handle);
}

template <class AddrT> LinuxSocketTraits<AddrT>* LinuxSocketTraits<AddrT>::accept(){
    struct sockaddr_in sa_addr;
    socklen_t len = sizeof(sa_addr);
    int ret = ::accept( m_handle, reinterpret_cast<struct sockaddr*>(&sa_addr), &len );
    if( ret < 0 && errno != EAGAIN ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    else if( ret < 0 && errno == EAGAIN ) return 0;
    return new LinuxSocketTraits( ret );
}

template <class AddrT> void LinuxSocketTraits<AddrT>::bind( const typename LinuxSocketTraits<AddrT>::AddrType& addr ){
    struct sockaddr_in sa_addr;
    addr >> &sa_addr;
    int ret = ::bind( m_handle, reinterpret_cast<struct sockaddr*>(&sa_addr), sizeof(sa_addr) );
    if( ret < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
}

template <class AddrT> void LinuxSocketTraits<AddrT>::connect( const typename LinuxSocketTraits<AddrT>::AddrType& addr ){
    struct sockaddr_in sa_addr;
    addr >> &sa_addr;
    int ret = ::connect( m_handle, reinterpret_cast<struct sockaddr*>(&sa_addr), sizeof(sa_addr));
    if( ret < 0 && errno != EINPROGRESS ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
}

template <class AddrT> typename LinuxSocketTraits<AddrT>::AddrType LinuxSocketTraits<AddrT>::getPeerName() const{
    struct sockaddr_in sa_addr;
    socklen_t len = sizeof(sa_addr);
    int ret = getpeername( m_handle, reinterpret_cast<struct sockaddr*>(&sa_addr), &len );
    if( ret < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    typename LinuxSocketTraits< AddrT >::AddrType addr;
    addr << &sa_addr;
    return addr;
}

template <class AddrT> typename LinuxSocketTraits<AddrT>::AddrType LinuxSocketTraits<AddrT>::getSockName() const{
    struct sockaddr_in sa_addr;
    socklen_t len = sizeof(sa_addr);
    int ret = getsockname( m_handle, reinterpret_cast<struct sockaddr*>(&sa_addr), &len);
    if( ret < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    typename LinuxSocketTraits< AddrT >::AddrType addr;
    addr << &sa_addr;
    return addr;
}

template <class AddrT> void LinuxSocketTraits<AddrT>::listen(int backlog){
    int ret=::listen( m_handle, backlog );
    if( ret < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
}

template <class AddrT> std::string LinuxSocketTraits<AddrT>::read(std::size_t length){
    char* buf(new char[length]);
    int ret = ::read( m_handle, buf, length );
    if( ret < 0 ){
        throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    }else if( ret == 0 ){
        throw SocketError( "Connection was closed by peer", typename SocketError::Error(Closed) );
    }
    std::string str(buf,ret);
    delete[] buf;
    return str;
}

template <class AddrT> std::string LinuxSocketTraits<AddrT>::read(){
    return read(bytesAvailable());
}

template <class AddrT> std::size_t LinuxSocketTraits<AddrT>::write(const std::string& str,std::size_t length){
    if( length > str.size() ) throw std::length_error("string length less that requested length to write");
    int ret = ::write( m_handle, str.c_str(), length );
    if( ret < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    return ret;
}

template <class AddrT> std::size_t LinuxSocketTraits<AddrT>::write(const std::string& str){
    return write( str, str.size() );
}

template <class AddrT> std::size_t LinuxSocketTraits<AddrT>::bytesAvailable() const{
    unsigned int nbytes = 0;
    if( ::ioctl( m_handle, FIONREAD, reinterpret_cast<char*>(&nbytes) ) < 0 ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
    return nbytes;
}


namespace {

template<class T> void addRemoveSocketHelper(std::list<T>& l1, std::list<T>& l2, const T& obj) {
	typename std::list<T>::iterator it = std::find(l2.begin(), l2.end(), obj);
	if(it == l2.end()) {
		l1.push_back(obj);
	} else {
		l2.erase(it);
	}	
}
template<class T> void dropRemovedSockets(std::list<T>& from, std::list<T>& what) {
	typedef typename std::list<T>::const_iterator const_iterator;
	typedef typename std::list<T>::iterator iterator;
	
	for(const_iterator it=what.begin(); it!=what.end(); ++it) {
		iterator it2 = std::find(from.begin(), from.end(), *it);
		if(it2 != from.end()) {
			from.erase(it2);
		}
	}

	what.clear();
}
template<class T> void pushAddedSockets(std::list<T>& to, std::list<T>& what) {
	typedef typename std::list<T>::const_iterator const_iterator;
	
	for(const_iterator it=what.begin(); it!=what.end(); ++it) {
		to.push_back(*it);
	}

	what.clear();
}

}

/*
  The issue here is that addSocket and removeSocket can be invocked from doPoll throught callbacks.
  So, we put object to the temporary queues to add or to remove.
*/
template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::addSocket(ClientSocket* client) {
	addRemoveSocketHelper(m_clientsToAdd, m_clientsToRemove, client);
}

template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::addSocket(ServerSocket* server) {
	addRemoveSocketHelper(m_serversToAdd, m_serversToRemove, server);
}

template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::removeSocket(ClientSocket* client) {
	addRemoveSocketHelper(m_clientsToRemove, m_clientsToAdd, client);
}

template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::removeSocket(ServerSocket* server) {
	addRemoveSocketHelper(m_serversToRemove, m_serversToAdd, server);
}

template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::flushSocketQueues() {
	dropRemovedSockets(m_servers, m_serversToRemove);
	pushAddedSockets(m_servers, m_serversToAdd);
	dropRemovedSockets(m_clients, m_clientsToRemove);
	pushAddedSockets(m_clients, m_clientsToAdd);
}

template <class AddrT> void LinuxSocketTraits<AddrT>::Watcher::doPoll( unsigned int time ){
    fd_set rfds, wfds, efds;
    struct timeval tv;
    int max_handle = -1;
    tv.tv_sec = time;
    tv.tv_usec = 0;

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);

    flushSocketQueues(); /* drop removed sockets and add new */

    for(typename std::list< ClientSocket* >::const_iterator ci = m_clients.begin(); ci != m_clients.end(); ++ci ) {
        ClientSocket* c = *ci;
        if( c->getState() == SocketBase::Idle ) continue;
        
        int handle = c->m_pImpl->m_handle;
        if( c->getState() == SocketBase::Connected ) {
            FD_SET(handle,&rfds);
            max_handle = std::max(max_handle, handle);
        }
        if( c->getState() != SocketBase::Idle ) {
            FD_SET(handle,&efds);
            max_handle = std::max(max_handle, handle);
        }
        if( c->getState() == SocketBase::Connecting ) {
            FD_SET(handle,&wfds);
            max_handle = std::max(max_handle, handle);
        }
    }

    for(typename std::list< ServerSocket* >::const_iterator si = m_servers.begin(); si != m_servers.end(); ++si ){
        ServerSocket* s = *si;
        if( s->getState() == SocketBase::Idle ) continue;
        
        int handle = s->m_pImpl->m_handle;
        if( s->getState() == SocketBase::Listening && s->hasAccepted() == 0 ){
            FD_SET(handle,&rfds);
            max_handle = std::max(max_handle, handle);
        }
        if( s->getState() != SocketBase::Idle ) {
            FD_SET(handle,&efds);
            max_handle = std::max(max_handle, handle);
        }
    }

    if( max_handle >= 0 ) {
        int ret = ::select( max_handle+1, &rfds, &wfds, &efds, &tv );
        if( ret < 0 ) {
            if( errno != EINTR ) throw SocketError( strerror(errno), typename SocketError::Error(errno) );
            else return;
        }
    } else {
        sleep( time );
        return;
    }
 
    int err = 0;
    socklen_t errlen = sizeof(err);

    for(typename std::list< ClientSocket* >::iterator ci = m_clients.begin(); ci != m_clients.end(); ++ci ){
        ClientSocket* c = *ci;
        if( c->getState() == SocketBase::Idle ) continue;

        int handle = c->m_pImpl->m_handle;
        if( FD_ISSET( handle, &efds ) ){
            getsockopt( handle, SOL_SOCKET, SO_ERROR, reinterpret_cast<void*>(&err), &errlen );
            c->onError( SocketError( strerror(err), typename SocketError::Error(err) ) );
        }
        if( FD_ISSET( handle, &wfds ) ){
            getsockopt( handle, SOL_SOCKET, SO_ERROR, reinterpret_cast<void*>(&err), &errlen );
            if( err == 0 )
                c->setState( SocketBase::Connected );
            else
                c->onError( SocketError( strerror(err), typename SocketError::Error(err) ) );
        }
        if( FD_ISSET( handle, &rfds ) ){
            getsockopt( handle, SOL_SOCKET, SO_ERROR, reinterpret_cast<void*>(&err), &errlen );
            if( err == 0) {
                char buf[1];
                int ret = ::recv( handle, buf, 1, MSG_PEEK|MSG_NOSIGNAL );
                if( ret == 0 ) {
                    c->close();
                } else if( ret < 0 )
                    c->onError( SocketError( strerror(errno), typename SocketError::Error(errno) ) );
                else
                    c->onDataRecieved();
            } else {
                c->onError( SocketError( strerror(err), typename SocketError::Error(err) ) );
            }
        }
    }
    
    for(typename std::list< ServerSocket* >::iterator si = m_servers.begin(); si != m_servers.end(); ++si ){
        ServerSocket* s = *si;
        if( s->getState() == SocketBase::Idle ) continue;

        int handle = s->m_pImpl->m_handle;
        if( FD_ISSET( handle, &efds ) ) {
            getsockopt( handle, SOL_SOCKET, SO_ERROR, reinterpret_cast<void*>(&err), &errlen );
            s->onError( SocketError( strerror(err), typename SocketError::Error(err) ) );
        }
        if( FD_ISSET( handle, &rfds) ) {
            s->onAccepted();
        }
    }
    
    flushSocketQueues(); /* drop removed sockets and add new */
}

template class LinuxSocketTraits<IP4Addr>;

